/**
 * @author doubledream
 * @desc 更新文件
 */

import * as fs from 'fs-extra'
import * as _ from 'lodash'
import * as prettier from 'prettier'
import * as ts from 'typescript'
import { readFile, writeFile, getSuffix } from './file'
import { getLangData } from './getLangData'
import { getProjectConfig, getLangDir, successInfo, failInfo, highlightText } from '../utils'

const CONFIG = getProjectConfig()
const srcLangDir = getLangDir(CONFIG.srcLang)

/**
 * 使用 Prettier 格式化文件
 * @param fileContent
 */
function prettierFile(fileContent) {
  try {
    return prettier.format(fileContent, {
      parser: 'typescript',
      trailingComma: 'all',
      singleQuote: true
    })
  } catch (e) {
    failInfo(`代码格式化报错！${e.toString()}\n代码为：${fileContent}`)
    return fileContent
  }
}

/**
 * 新建一个新的语言文件
 * @param key 语言文件中的I80N值
 * @param value 语言文件中的I18N值
 */
function generateNewLangFile(key, value) {
  const obj = _.set({}, key, value)

  return prettierFile(`export default ${JSON.stringify(obj, null, 2)}`)
}

/**
 * 在语言index文件中新增语言文件引入
 * @param newFilename 新的语言文件名
 */
function addImportToMainLangFile(newFilename) {
  const indexFileSrc = `${srcLangDir}/index.${CONFIG.isJsProj ? 'js' : 'ts'}`
  let mainContent = ''
  // 如果已经存在index文件则更新内容
  if (fs.existsSync(indexFileSrc)) {
    mainContent = fs.readFileSync(indexFileSrc, 'utf8')
    mainContent = mainContent.replace(/^(\s*import.*?)$/m, `$1\nimport ${newFilename} from './${newFilename}'`)
    if (/(}\))/.test(mainContent)) {
      if (/\,\n(}\))/.test(mainContent)) {
        // 最后一行包含,号
        mainContent = mainContent.replace(/(}\))/, `  ${newFilename},\n$1`)
      } else {
        // 最后一行不包含,号
        mainContent = mainContent.replace(/\n(}\))/, `,\n  ${newFilename},\n$1`)
      }
    }
    // 兼容 export default { common }的写法
    if (/(})/.test(mainContent)) {
      if (/\,\n(})/.test(mainContent)) {
        // 最后一行包含,号
        mainContent = mainContent.replace(/(})/, `  ${newFilename},\n$1`)
      } else {
        // 最后一行不包含,号
        mainContent = mainContent.replace(/\n(})/, `,\n  ${newFilename},\n$1`)
      }
    }
  }
  // 还未存在index文件，生成初始值
  else {
    mainContent = `import ${newFilename} from './${newFilename}'\n\nexport default Object.assign({}, {\n  ${newFilename},\n})`
  }

  // 写入文件
  fs.writeFileSync(indexFileSrc, mainContent)
}

/**
 * 更新语言文件
 * @param key 代码文件中的I18N键
 * @param val 语言文件中的I80N值
 * @param validateDuplicate 是否校验重复
 */
function updateLangFiles(key, val, validateDuplicate) {
  // 过滤掉不是I18N开头的Key
  if (!_.startsWith(key, 'I18N.')) {
    return
  }
  // 拆分代码文件中的键
  const [, filename, ...restPath] = key.split('.')
  // 语言文件中的I18N键
  const fullKey = restPath.join('.')
  // 语言文件路径
  const langFilePath = `${srcLangDir}/${filename}.${CONFIG.isJsProj ? 'js' : 'ts'}`

  // 还未生成对应的语言文件
  if (!fs.existsSync(langFilePath)) {
    // 生成新的语言文件内容并写入
    fs.writeFileSync(langFilePath, generateNewLangFile(fullKey, val))
    // 在index文件中增加语言文件引入
    addImportToMainLangFile(filename)
    successInfo(`成功新建语言文件 ${langFilePath}`)
  }
  else {
    // 清除 require 缓存，解决手动更新语言文件后再自动抽取，导致之前更新失效的问题
    const mainContent = getLangData(langFilePath)

    // 解析失败会返回{}，需要过滤下
    if (Object.keys(mainContent).length === 0) {
      failInfo(`${filename} 解析失败，该文件包含的文案无法自动补全`)
    }

    // 校验是否有重复的Key
    if (validateDuplicate && _.get(mainContent, fullKey) !== undefined) {
      failInfo(`${langFilePath} 中已存在 key 为 \`${fullKey}\` 的翻译，请重新命名变量`)
      throw new Error('duplicate')
    }
    // \n 会被自动转义成 \\n，这里转回来
    val = val.replace(/\\n/gm, '\n')
    // 写入语言文件
    _.set(mainContent, fullKey, val)
    fs.writeFileSync(langFilePath, prettierFile(`export default ${JSON.stringify(mainContent, null, 2)}`))
  }
}

/**
 * 检查代码文件是否添加过 import I18N 命令
 * @param filePath 文件路径
 */
function hasImportI18N(filePath) {
  const code = readFile(filePath)
  const ast = ts.createSourceFile('', code, ts.ScriptTarget.ES2015, true, ts.ScriptKind.TSX)
  let hasImportI18N = false

  function visit(node) {
    if (node.kind === ts.SyntaxKind.ImportDeclaration) {
      const importClause = node.importClause

      // import I18N from 'src/utils/I18N'
      if (_.get(importClause, 'kind') === ts.SyntaxKind.ImportClause) {
        if (importClause.name) {
          if (importClause.name.escapedText === 'I18N') {
            hasImportI18N = true
          }
        } else {
          const namedBindings = importClause.namedBindings
          // import { I18N } from 'src/utils/I18N'
          if (namedBindings.kind === ts.SyntaxKind.NamedImports) {
            namedBindings.elements.forEach(element => {
              if (element.kind === ts.SyntaxKind.ImportSpecifier && _.get(element, 'name.escapedText') === 'I18N') {
                hasImportI18N = true
              }
            })
          }
          // import * as I18N from 'src/utils/I18N'
          if (namedBindings.kind === ts.SyntaxKind.NamespaceImport) {
            if (_.get(namedBindings, 'name.escapedText') === 'I18N') {
              hasImportI18N = true
            }
          }
        }
      }
    }
  }

  ts.forEachChild(ast, visit)
  return hasImportI18N
}

/**
 * 在合适的位置添加 import I18N 语句
 * @param filePath 文件路径
 */
function createImportI18N(filePath) {
  const code = readFile(filePath)
  const ast = ts.createSourceFile('', code, ts.ScriptTarget.ES2015, true, ts.ScriptKind.TSX)
  const isNomalFile = ['ts', 'js', 'tsx', 'jsx'].includes(getSuffix(filePath))
  const isVueFile = _.endsWith(filePath, '.vue')
  // 正常文件开头添加import
  if (isNomalFile) {
    const importStatement = `${CONFIG.importI18N}\n`
    const pos = ast.getStart(ast, false)
    const updateCode = code.slice(0, pos) + importStatement + code.slice(pos)

    return updateCode
  }
  // Vue文件开头添加import
  else if (isVueFile) {
    const importStatement = `${CONFIG.importI18N}\n`
    const updateCode = code.replace(/<script>/g, `<script>\n${importStatement}`)
    return updateCode
  }
}

/**
 * 更新语言文件以及源代码
 * @param filePath 当前代码文件路径
 * @param target  目标字符串对象
 * @param key  目标 key
 * @param validateDuplicate 是否校验文件中已经存在要写入的 key
 * @param needWrite 是否只需要替换不需要更新 langs 文件
 */
function replaceAndUpdate(filePath, target, key, validateDuplicate, needWrite = true) {
  // 获取目标字段内容
  const { text, range: { start, end }, isString } = target
  // 获取当前代码内容
  const code = readFile(filePath)
  const isHtmlFile = _.endsWith(filePath, '.html')
  const isVueFile = _.endsWith(filePath, '.vue')
  // 写回代码文件的内容
  let newCode = code
  // 中文语言结果
  let cnVal = text

  // 若是字符串，删掉两侧的引号
  if (isString) {
    // 写回代码文件的I18N键
    let codeKey = key
    // 如果引号左侧是 等号，则可能是 jsx 的 props，此时要替换成 {
    const preTextStart = start - 1
    const [last2Char, last1Char] = code.slice(preTextStart, start + 1).split('')
    if (last2Char === '=') {
      if (isHtmlFile) {
        codeKey = '{{' + key + '}}'
      } else if (isVueFile) {
        codeKey = '{{' + key + '}}'
      } else {
        codeKey = '{' + key + '}'
      }
    }
    // 若是模板字符串，看看其中是否包含变量
    if (last1Char === '`') {
      // 正则可视化 https://c.runoob.com/front-end/7625/#!flags=&re=(%5C%24%5C%7B%5B%5E%5C%7D%5D%2B%3F%5C%7D)
      // 匹配模板字符串内的全部变量内容
      const varInStr = text.match(/(\$\{[^\}]+?\})/g)
      // 如果存在变量
      if (varInStr) {
        /**
         * 取出变量内容并用val{x}指代
         * 示例：[ 'val1: a', "val2: '全球每个人'" ]
         */
        const kvPair = varInStr.map((str, index) => {
          return `val${index + 1}: ${str.replace(/^\${([^\}]+)\}$/, '$1')}`
        })
        /**
         * 包装模板字符串的I18N替换代码，用于写回代码文件
         * 示例：I18N.template(I18N.home.index.shiZhongJianChiZuo, { val1: a, val2: '全球每个人' })
         */
        codeKey = `I18N.template(${key}, { ${kvPair.join(',\n')} })`
        /**
         * 使用val{x}替换掉原模板字符串的变量，用于生成语言文件
         * 示例：始终坚持做“{val1}”的好产品，让{val2}都能享受科技带来的美好生活
         */
        varInStr.forEach((str, index) => {
          cnVal = cnVal.replace(str, `{val${index + 1}}`)
        })
      }
    }
    // 将I18N替换代码写回代码文件
    newCode = `${code.slice(0, start)}${codeKey}${code.slice(end)}`
  } else {
    if (isHtmlFile || isVueFile) {
      newCode = `${code.slice(0, start)}{{${key}}}${code.slice(end)}`
    } else {
      newCode = `${code.slice(0, start)}{${key}}${code.slice(end)}`
    }
  }
  try {
    if (needWrite) {
      // 更新语言文件
      updateLangFiles(key, cnVal, validateDuplicate)
    }
    // // 若更新成功再替换代码
    return writeFile(filePath, newCode)
  } catch (e) {
    return Promise.reject(e.message)
  }
}

export { replaceAndUpdate, hasImportI18N, createImportI18N }
